@@ -14,9 +14,12 @@
 
 import os
 import dbus
+import registry
+import string
 
 from interfaces import *
-from role import ROLE_INVALID
+from role import ROLE_DESKTOP_FRAME
+import state
 
 from busutils import *
 
@@ -25,9 +28,9 @@ __all__ = [
            "AccessibleCache"
           ]
 
-_ATSPI_CACHE_PATH = '/org/at_spi/cache'
-_ATSPI_CACHE_INTERFACE = 'org.freedesktop.atspi.Cache'
-_ATSPI_EVENT_OBJECT_INTERFACE = "org.freedesktop.atspi.Event.Object"
+_ATSPI_CACHE_PATH = '/org/a11y/atspi/cache'
+_ATSPI_CACHE_INTERFACE = 'org.a11y.atspi.Cache'
+_ATSPI_EVENT_OBJECT_INTERFACE = "org.a11y.atspi.Event.Object"
 
 #------------------------------------------------------------------------------
 
@@ -42,6 +45,7 @@ class _CacheData(object):
                         'name',
                         'description',
                         'state',
+                        'extraData'
                     ]
 
         def __init__(self, data):
@@ -86,8 +90,6 @@ class DesktopCacheManager (object):
                 self._application_list = {}
 
                 bus.add_signal_receiver(self._children_changed_handler,
-                                        #bus_name=ATSPI_REGISTRY_NAME,
-                                        #path=ATSPI_ROOT_PATH,
                                         dbus_interface=_ATSPI_EVENT_OBJECT_INTERFACE,
                                         signal_name="ChildrenChanged",
                                         interface_keyword="interface",
@@ -95,6 +97,24 @@ class DesktopCacheManager (object):
                                         sender_keyword="sender",
                                         path_keyword="path")
 
+                self._property_change =  \
+                        bus.add_signal_receiver(self._property_change_handler,
+                                                dbus_interface=_ATSPI_EVENT_OBJECT_INTERFACE,
+                                                signal_name="PropertyChange",
+                                                interface_keyword="interface",
+                                                member_keyword="member",
+                                                sender_keyword="sender",
+                                                path_keyword="path")
+
+                self._state_changed = \
+                        bus.add_signal_receiver(self._state_changed_handler,
+                                                dbus_interface=_ATSPI_EVENT_OBJECT_INTERFACE,
+                                                signal_name="StateChanged",
+                                                interface_keyword="interface",
+                                                member_keyword="member",
+                                                sender_keyword="sender",
+                                                path_keyword="path")
+
                 obj     = bus.get_object(ATSPI_REGISTRY_NAME, ATSPI_ROOT_PATH, introspect=False)
                 desktop = dbus.Interface(obj, ATSPI_ACCESSIBLE)
                 apps    = desktop.GetChildren()
@@ -111,7 +131,7 @@ class DesktopCacheManager (object):
                                        apps,                                    #Children 
                                        [ATSPI_ACCESSIBLE, ATSPI_COMPONENT],     #Interfaces
                                        "main",                                  #Name
-                                       ROLE_INVALID,                            #Role
+                                       ROLE_DESKTOP_FRAME,                            #Role
                                        "",                                      #Description
                                        [0,0]                                    #State
                                      )
@@ -120,26 +140,46 @@ class DesktopCacheManager (object):
                 for bus_name, object_path in apps:
                                 self._application_list[bus_name] = ApplicationCacheManager (cache, bus_name)
 
-	def _children_changed_handler (self, 
-                                       app, minor, detail1, detail2, any_data,
-				       interface=None, sender=None, member=None, path=None):
+        def _children_changed_handler (self, 
+                                       minor, detail1, detail2, any_data, app,
+                                       interface=None, sender=None, member=None, path=None):
                 if interface==_ATSPI_EVENT_OBJECT_INTERFACE and sender == self._unique_name and path == ATSPI_ROOT_PATH:
-		        if minor == "add":
+                        if minor == "add":
                                 bus_name, object_path = any_data
                                 self._application_list[bus_name] = ApplicationCacheManager(self._cache, bus_name)
-		        elif minor == "remove":
+                                self._cache[(self._unique_name, ATSPI_ROOT_PATH)].children.append (any_data)
+                        elif minor == "remove":
                                 bus_name, object_path = any_data
                                 self._application_list[bus_name].remove_all()
                                 del(self._application_list[bus_name])
+                                self._cache[(self._unique_name, ATSPI_ROOT_PATH)].children.remove (any_data)
 
-			item = self._cache[(sender, path)]
-			if minor == "add":
-				item.children.insert (detail1, any_data)
-			elif minor == "remove":
-				del (item.children[detail1])
+                        #item = self._cache[(sender, path)]
+                        #if minor == "add":
+                        #        item.children.insert (detail1, any_data)
+                        #elif minor == "remove":
+                        #        del (item.children[detail1])
+
+                if sender in self._application_list:
+                        app = self._application_list[sender]
+                        app._children_changed_handler(minor, detail1, detail2, any_data, app, interface, sender, member, path)
 
 #------------------------------------------------------------------------------
 
+        def _property_change_handler (self,
+                                       minor, detail1, detail2, any_data, app,
+                                       interface=None, sender=None, member=None, path=None):
+                if sender in self._application_list:
+                        app = self._application_list[sender]
+                        app._property_change_handler(minor, detail1, detail2, any_data, app, interface, sender, member, path)
+
+        def _state_changed_handler (self,
+                                       minor, detail1, detail2, any_data, app,
+                                       interface=None, sender=None, member=None, path=None):
+                if sender in self._application_list:
+                        app = self._application_list[sender]
+                        app._state_changed_handler(minor, detail1, detail2, any_data, app, interface, sender, member, path)
+
 class ApplicationCacheManager (object):
         """
         The application cache manager is responsible for keeping the cache up to date
@@ -159,40 +199,23 @@ class ApplicationCacheManager (object):
                 bus = AsyncAccessibilityBus()
 
                 self._cache = cache
-		self._bus_name = bus_name
+                self._bus_name = bus_name
 
                 cache_obj = bus.get_object (bus_name, _ATSPI_CACHE_PATH, introspect=False)
                 cache_itf = dbus.Interface (cache_obj, _ATSPI_CACHE_INTERFACE)
-                self._add_objects(cache_itf.GetItems())
+                try:
+                        self._add_objects(cache_itf.GetItems())
+                except dbus.exceptions.DBusException:
+                        pass
 
-                self._property_change =  \
-                        bus.add_signal_receiver(self._property_change_handler,
-                                                bus_name=self._bus_name,
-                                                dbus_interface=_ATSPI_EVENT_OBJECT_INTERFACE,
-                                                signal_name="PropertyChange",
-                                                interface_keyword="interface",
-                                                member_keyword="member",
-                                                sender_keyword="sender",
-                                                path_keyword="path")
-
-                self._children_changed = \
-                        bus.add_signal_receiver(self._children_changed_handler,
-                                                bus_name=self._bus_name,
-                                                dbus_interface=_ATSPI_EVENT_OBJECT_INTERFACE,
-                                                signal_name="ChildrenChanged",
-                                                interface_keyword="interface",
-                                                member_keyword="member",
-                                                sender_keyword="sender",
-                                                path_keyword="path")
-
-		self._cache_add = \
+                self._cache_add = \
                         bus.add_signal_receiver(self._add_object,
                                                 bus_name=self._bus_name,
                                                 path=_ATSPI_CACHE_PATH,
                                                 dbus_interface=_ATSPI_CACHE_INTERFACE,
                                                 signal_name="AddAccessible")
 
-		self._cache_remove = \
+                self._cache_remove = \
                         bus.add_signal_receiver(self._remove_object,
                                                 bus_name=self._bus_name,
                                                 path=_ATSPI_CACHE_PATH,
@@ -206,36 +229,61 @@ class ApplicationCacheManager (object):
                 self._cache[(bus_name, object_path)] = _CacheData (data)
 
         def _remove_object(self, reference):
-		bus_name, object_path = reference
-                del(self._cache[(bus_name, object_path)])
+                bus_name, object_path = reference
+                try:
+                        del(self._cache[(bus_name, object_path)])
+                except KeyError:
+                        pass
 
         def _add_objects (self, objects):
                 for data in objects:
                         self._add_object (data)
 
-	def _property_change_handler (self,
-                                      app, minor, detail1, detail2, any_data,
-				      interface=None, sender=None, member=None, path=None):
+        def _property_change_handler (self,
+                                      minor, detail1, detail2, any_data, app,
+                                      interface=None, sender=None, member=None, path=None):
                 if interface==_ATSPI_EVENT_OBJECT_INTERFACE:
                         if (sender, path) in self._cache:
-			        item = self._cache[(sender, path)]
-			        if minor == "accessible-name":
-				        item.name = any_data
-			        elif minor == "accessible-description":
-				        item.description = any_data
-			        elif minor == "accessible-parent":
-				        item.parent = any_data
+                                item = self._cache[(sender, path)]
+                                if minor == "accessible-name":
+                                        item.name = any_data
+                                if minor == "accessible-role":
+                                        item.role = any_data
+                                elif minor == "accessible-description":
+                                        item.description = any_data
+                                elif minor == "accessible-parent":
+                                        item.parent = any_data
 
-	def _children_changed_handler (self,
-                                       app, minor, detail1, detail2, any_data,
-				       interface=None, sender=None, member=None, path=None):
+        def _children_changed_handler (self,
+                                       minor, detail1, detail2, any_data, app,
+                                       interface=None, sender=None, member=None, path=None):
                 if interface==_ATSPI_EVENT_OBJECT_INTERFACE:
-		        if (sender, path) in self._cache:
-			        item = self._cache[(sender, path)]
-			        if minor == "add":
-				        item.children.insert (detail1, any_data)
-			        elif minor == "remove":
-				        del (item.children[detail1])
+                        if (sender, path) in self._cache:
+                                item = self._cache[(sender, path)]
+                                if item.state[0] & (1 << state.STATE_MANAGES_DESCENDANTS):
+                                        return
+                                if minor.startswith("add"):
+                                        item.children.insert (detail1, any_data)
+                                elif minor.startswith("remove"):
+                                        item.children.remove (any_data)
+                                        if any_data in self._cache:
+                                                child = self._cache[any_data]
+                                                if child.parent == (sender, path):
+                                                        child.parent = (sender, ATSPI_NULL_PATH)
+
+        def _state_changed_handler (self,
+                                       minor, detail1, detail2, any_data, app,
+                                       interface=None, sender=None, member=None, path=None):
+                if interface==_ATSPI_EVENT_OBJECT_INTERFACE:
+                        if (sender, path) in self._cache:
+                                item = self._cache[(sender, path)]
+                                val = eval("int(state.STATE_" + string.upper(minor) + ")")
+                                high = int(val / 32)
+                                low = val % 32
+                                if (detail1 == 1):
+                                        item.state[high] |= (1 << low)
+                                else:
+                                        item.state[high] &= ~(1 << low)
 
         def remove_all (self):
                 for bus_name, object_path in self._cache.keys():
@@ -246,8 +294,8 @@ class ApplicationCacheManager (object):
 
 class AccessibleCache (dict):
 
-	def __init__ (self, bus_name=None):
-		dict.__init__ (self)
+        def __init__ (self, bus_name=None):
+                dict.__init__ (self)
 
                 if bus_name:
                         self._manager = ApplicationCacheManager (self, bus_name) 
